/*

  Schiffe versenken fuer ATmega8 Ping-Pong-Spiel
  Autor: Klaus Peichl (pei@freenet.de)
  Datum: 2010-03-31

  Compilierung:
  avr-gcc -mmcu=atmega8 schiffe.c
  avr-objcopy -j .text -O ihex a.out schiffe.hex
  wine ISPmega8.exe

  Linux-Target:
  gcc schiffe.c -lncurses -g -Wall -o schiffe

  History:

  Version    Beschreibung
  -----------------------------------------------------------------------------------
  1.0        "Feature complete"

  1.1        Behobene Fehler:
             - Das Zielraster des Computergegners übersprang die linke Spalte
	       und war dadurch auch anderso nicht engmaschig genug, so dass
	       das Spiel hängenbleiben konnte. Außerdem wurde dadurch die
	       Spielstärke des Computergegners beeinträchtigt.
	     - Nach einem gewonnenem Spiel war das nächste Spiel sofort wieder gewonnen.
	     - Das Spiel schaltete sich nicht ab und der Münzkontakt war ohne Funktion.
	     - Unnötiges Scrollen, wenn der Computer zuerst schießt.
	     - Ständiges Hin- und Herdrehen beim Plazieren eines Schiffes in der
	       linken oberen oder unteren Ecke.
	     - Die Variable "randomstate" musste als volatile deklariert werden, damit
	       flackernde LEDs in den obersten 2 Reihen nicht nur langsam blinken.
	     - Flackernde LEDs waren während eines Schusses eingefroren, weil
	       randomize() nicht aufgerufen wurde.


Ports fuer LED-Zeilen-Ansteuerung:
  C0
  C1
  C2
  C3
  D4
  D5
  D6
  D7
  B0
  B1

Port fuer Soundausgabe:
  B7


Endbildschirm:

                1
x x  x  x x     2
x x x x x x     4
 x  x x x x     8
 x   x  xxx    16
               32
x   x x x  x   64
x x x x xx x  128
x x x x x xx  256
 x x  x x  x  512

                1
     xxx        2
      x         4
      x         8
     xxx       16
               32
x   x x x  x   64
x x x x xx x  128
x x x x x xx  256
 x x  x x  x  512

*/


typedef unsigned short BITMAP[2][12];

#ifdef __AVR__
volatile static BITMAP vram;
#define VRAM vram
#else
BITMAP * pVram;
#define VRAM (*pVram)
#endif


static BITMAP map1; /* Eigenes Seegebiet */
static BITMAP map2; /* Gegnerisches Seegebiet */
static BITMAP map3; /* Karte des Computerspielers */
static BITMAP save;


static BITMAP endscreen = {
  { 448, 512, 384, 512, 448, 18, 990, 18, 960, 128, 256, 960 },
  { 454, 536, 390, 512, 460, 18, 972, 0, 990, 144, 286, 960 }
};


/* Position des Spielfelds horizontal zentriert */
#define XMIN 1
#define XMAX 10
#define YMIN 0
#define YMAX 9

/* Anzahl der Schiffe je Spieler */
#define SHIPS 5

/*
  Position, Lage und Zustand jedes Schiffs werden in 16 Bit gespeichert:
  L L V Z Z Z Z Z  Y Y Y Y X X X X

  L = Laenge - 2
  V = Vertikal-Flag
  Z = Zustand pro Segment (1=unbeschaedigt)
  Y = Position
  X = Position
*/
static unsigned short ships[2*SHIPS];
volatile static unsigned short randomstate;

static char targetx = -1;
static char targety = -1;
static char targetsize;
static char rasterx = 3;
static char rastery = 6;
static char targethits;
static char totalhits;
static char playerhits;
static char maxhits;
static char flanke[4];
static char richtung;
static char distanz;


#ifdef __AVR__

#include <avr/io.h>         
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>


/*
  Interrupt-Handler für Münzkontakt
*/
ISR(INT0_vect)
{
  /* INT0 abschalten */
  GICR &= (~0x40);
}


void WaitCoin(void)
{
  /* Timer anhalten */
  TCCR0 = 0;

  /* ADC abschalten */
  ADCSRA = 0;

  PORTB = 0;
  PORTC = 0;
  PORTD = 0;

  /* PORTD4 für Münzkontakt auf INPUT schalten und als INT0 verwenden */
  DDRD &= 251; 
  GICR |= 0x40;

  /* Sparsamsten Schlafmodus wählen (ca. 0.5 µA) und aktivieren */
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_mode();
}


static volatile int  col = 0;
static volatile int  phase = 0;


/*
  LED-Multiplexing-Interrupt
*/
ISR(TIMER0_OVF_vect)
{

  int led0, led1, leds;
  int layer;

  if (col++ > 11) {
    col = 0;
    PORTB &= ~16;
    phase++;
    phase = phase & 7; 
    /* phase & 3: Ruhiges Bild, aber wenig Kontrast
       phase & 7: Besserer Kontrast, aber Flimmern */
  }
  else {
    PORTB |= 16;
  }

  /* LEDs abschalten */
  PORTB &= ~0x03;
  PORTC &= ~0x0F;
  PORTD &= ~0xF0;

  /* CLK-Impuls */
  PORTB |= 8;
  PORTB &= ~8;

  /* STR-Impuls */
  PORTB |= 4;
  PORTB &= ~4;

  /* Anodenleitungen gemaeß Muster aktivieren */
  led0 = VRAM[0][col];
  led1 = VRAM[1][col];

  led1 = (led1 & randomstate) | (led0 & led1);

  switch (phase) {
  case 0:
    leds = led0;
    break;
  case 1:
  case 2:
    leds = led1;
    break;
  default:
    leds = led0 & led1;
    break;
  }

  PORTB |= (leds >> 8 ) & 0x03;
  PORTC |= leds & 0x0F;
  PORTD |= leds & 0xF0;
}


int GetADChannel(int channel) {
  ADMUX = (1<<REFS0) | channel;
  ADCSRA |= (1<<ADSC);
  while ( ADCSRA & (1<<ADSC) );
  return ADC;
}

#else

#include <ncurses.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>


FILE * dspfile;

void RefreshScreen()
{
  int x,y;

  init_pair(1, COLOR_BLACK, COLOR_BLACK);
  init_pair(2, COLOR_BLUE,  COLOR_BLUE);
  init_pair(3, COLOR_RED,   COLOR_RED);
  init_pair(4, COLOR_WHITE, COLOR_WHITE);

  for (x=0; x<12; x++) {
    unsigned short led0 = VRAM[0][x];
    unsigned short led1 = VRAM[1][x];
    led1 = (led1 & randomstate) | (led0 & led1);

    for (y=0; y<10; y++) {
      int color = ((led0 >> y) & 1) + (((led1 >> y) & 1 )<< 1) ;
      attrset(COLOR_PAIR(color+1));
      mvprintw( y, x+1,".");
    }
  }
  randomize();
  refresh();
}
#endif


static void Init(void)
{
#ifdef __AVR__
  cli();

  DDRB= 0xFF;
  DDRC= 0x0F;
  DDRD= 0xFF;

  /* Timer-Interrupt */
  TCCR0= (1<<CS01);
  TIMSK= (1<<TOIE0);

  sei();

  /* ADC */
  ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);
  ADMUX = (1<<REFS0);

#else

  /* Shared memory */
  pid_t forkid;
  key_t key;
  int shmid;


  /* ncurses window */
  char ch;
  WINDOW * wnd = initscr();
  clear();
  noecho();
  nodelay(wnd, 1);
  cbreak();
  start_color();
  ch = getch();


  /* make shared memory key: */
  if ((key = ftok("/tmp/schiffe-shm", 'R')) == -1) {
    perror("ftok");
    exit(1);
  }

  /* connect to (and possibly create) the segment: */
  if ((shmid = shmget(key, sizeof(BITMAP), 0644 | IPC_CREAT)) == -1) {
    perror("shmget");
    exit(1);
  }

  signal(SIGCHLD, SIG_IGN); /* avoid stale child process */
  forkid = fork();

  switch ( forkid ) {
  case -1:
    fprintf(stderr, "Could not fork.\n");
    exit(EXIT_FAILURE);
  case 0:
    /* Pointer auf Video-RAM im Kind-Prozess */
    pVram = shmat(shmid, (void *)0, 0);
    if (pVram == (char *)(-1)) {
      perror("shmat");
      exit(1);
    }
    while (1) {
      /* Simulation des Timer-Interrupts für die LED-Matrix */
      RefreshScreen();
      /* Kurze Pause zur Vermeidung von hoher CPU-Last */
      delay(30);
    }
  default:
    /* Pointer auf Video-RAM im Mutter-Prozess */
    pVram = shmat(shmid, (void *)0, 0);
    if ((char *)pVram == (char *)(-1)) {
      perror("shmat");
      exit(1);
    }
  }
#endif
}


void randomize(void)
{
  randomstate = (0x0529 * randomstate) + 0x3A7F;
}


void delay(int millisec)
{
#ifdef __AVR__
  int i, j;
  for (i=0; i<millisec; i++) {
    for (j=0; j<260; j++) {
      randomize();
    }
  }
#else
  usleep(1000 * millisec);
#endif
}


void SpeakerOn(int duration)
{
  int i;
#ifdef __AVR__
  for (i=0; i<duration; i++)
    PORTB |= 128;
#else
  unsigned char sample = 128;
  for (i=0; i<duration>>8; i++) {
    if (dspfile) {
      fwrite(&sample, 1, 1, dspfile);
    }
  }
  usleep(duration>>2);
#endif
}


void SpeakerOff(int duration)
{
  int i;
#ifdef __AVR__
  for (i=0; i<duration; i++)
    PORTB &= 127;
#else
  unsigned char sample = 0;
  for (i=0; i<duration>>8; i++) {
    if (dspfile) {
      fwrite(&sample, 1, 1, dspfile);
    }
  }
  usleep(duration>>2);
#endif
}


void click(void)
{
  int i;
#ifndef __AVR__
  dspfile = fopen("/dev/dsp", "wb");
#endif

  for (i=0; i<20; i++) {
    SpeakerOn(200);
    SpeakerOff(200);
  }

#ifndef __AVR__
  if (dspfile) {
    fclose(dspfile);
  }
#endif
}


void Beep(int dt)
{
  int t = 0;
#ifndef __AVR__
  dspfile = fopen("/dev/dsp", "wb");
#endif

  while (t < 20000) {
    SpeakerOn(dt);
    SpeakerOff(dt);
    t += dt;
  }

#ifndef __AVR__
  if (dspfile) {
    fclose(dspfile);
  }
#endif
}


void flipsound() {
  Beep(2000);
  Beep(500);
}


void ClearScreen(BITMAP dest)
{
  int x;
  for (x=0; x<12; x++) {
    dest[0][x] = 0;
    dest[1][x] = 0;
  }
}

void CopyScreen(BITMAP src, BITMAP dest)
{
  int x;
  for (x=XMIN; x<=XMAX; x++) {
    dest[0][x] = src[0][x];
    dest[1][x] = src[1][x];
  }
}


void ScrollDownScreen(BITMAP src)
{
  int x, y;
  for (y=9; y>=0; y--) {
    for (x=XMIN; x<=XMAX; x++) {
      VRAM[0][x] = (VRAM[0][x] << 1) | (src[0][x] >> y);
      VRAM[1][x] = (VRAM[1][x] << 1) | (src[1][x] >> y);
    }
    delay(100);
  }
}


void ScrollUpScreen(BITMAP src)
{
  int x, y;
  for (y=9; y>=0; y--) {
    for (x=XMIN; x<=XMAX; x++) {
      VRAM[0][x] = (VRAM[0][x] >> 1) | ((src[0][x] << y) & 0x0200);
      VRAM[1][x] = (VRAM[1][x] >> 1) | ((src[1][x] << y) & 0x0200);
    }
    delay(100);
  }
}


void PutScreenPixel(int x, int y, BITMAP screen, int color)
{
  if ( ((unsigned int)x < 12) && ((unsigned int)y < 10) ) {
    int bit   = (color & 1) << y;
    int mask  = ~(1 << y);
    screen[0][x] = (screen[0][x] & mask) | bit;
    bit  = (color >> 1) << y;
    screen[1][x] = (screen[1][x] & mask) | bit;
  }
}


void PutPixel(int x, int y, int color)
{
  PutScreenPixel(x, y, VRAM, color);
}


int GetScreenPixel(int x, int y, BITMAP screen)
{
  if ( ((unsigned int)x < 12) && ((unsigned int)y < 10) ) {
    return ((screen[0][x] >> y) & 1) + (((screen[1][x] >> y) & 1) << 1);
  }
  else {
    return 0;
  }
}


int GetPixel(int x, int y)
{
  return GetScreenPixel(x, y, VRAM);
}


void MergePixel(int x, int y, BITMAP src)
{
  int oldcolor = GetScreenPixel(x, y, src);
  int color;

  switch (oldcolor) {
  case 0:
    /* Fadenkreuz alleine */
    color = 1;
    break;
  case 1:
    /* Frühere Ziele werden heller */
  case 3:
    /* Schiffe bleiben under Fadenkreuz sichtbar */
    color = 3;
    break;
  default:
    /* Feuer bleibt unverändert */
    color = 2;
  }
    
  PutPixel(x, y, color);
}


void FlashPixel(int x, int y)
{
  int i, color;

  color = GetPixel(x, y);

  for (i=0; i<4; i++) {
    PutPixel(x, y, 3);
    delay(18);
    PutPixel(x, y, 0);
    delay(18);
  }
  PutPixel(x, y, color);
}

   
void HLineScreen(int x, int y, int len, BITMAP screen, int color)
{
  for ( ; len>0; len-- )
    PutScreenPixel(x++, y, screen, color);
}


void HLine(int x, int y, int len, int color)
{
  HLineScreen(x, y, len, VRAM, color);
}


void VLineScreen(int x, int y, int len, BITMAP screen, int color)
{
  for ( ; len>0; len-- )
    PutScreenPixel(x, y++, screen, color);
}


void VLine(int x, int y, int len, int color)
{
  VLineScreen(x, y, len, VRAM, color);
}


void TimeBar(int timer)
{
  /* Restzeit-Balken */
  int len = (timer >> 4);
  VLine(0,    0,  10-len, 0);
  VLine(0, 10-len, len,   1);
  VLine(11,    0,  10-len, 0);
  VLine(11, 10-len, len,   1);
}


/*
  stdcolor:   0-3: Helligkeit für intakte Segmente
               -1: Schiff unsichtbar
  wreckcolor: 0-3: Helligheit des Wracks
*/
int DrawShip(unsigned short ship, BITMAP screen, int stdcolor, int wreckcolor)
{
  int size = (ship >> 14) + 2;
  int x = ship & 0x000F;
  int y = (ship >> 4) & 0x000F;
  int color, i;
  int uebrig = 0;

  for (i=0; i<size; i++) {
    if (ship & 0x1F00) {
      if (ship & (0x100 << i)) {
	/* Segment unbeschädigt */
	color = stdcolor;
	uebrig++;
      }
      else {
	/* Segment brennt */
	color = 2;
      }
    }
    else {
      /* Schiff versenkt */
      color = wreckcolor;
    }

    if (color != -1) {
      if (ship & 0x2000)
	PutScreenPixel(x, y+i, screen, color);
      else
	PutScreenPixel(x+i, y, screen, color);
    }
  }

  return uebrig;
}


void GetXY(int *px, int*py) {
  static int quantx = 4;
  static int quanty = 4;
#ifdef __AVR__
  static int adx = 500;
  static int ady = 500;
  int tmp, delta;
  const int hyst = 20;

  /* Glaettung */
  adx = (adx + GetADChannel(7)) >> 1;
  ady = (ady + GetADChannel(6)) >> 1;

  /* Hysterese */
  tmp = 13 - (adx >> 6);
  delta = tmp - quantx;
  if (delta == 1) {
    tmp = 13 - ((adx-hyst) >> 6);
  }
  if (delta == -1) {
    tmp = 13 - ((adx+hyst) >> 6);
  }
  quantx = tmp;

  tmp = -2 + (ady >> 6);
  delta = tmp - quanty;
  if (delta == 1) {
    tmp = -2 + ((ady-hyst) >> 6);
  }
  if (delta == -1) {
    tmp = -2 + ((ady+hyst) >> 6);
  }
  quanty = tmp;
#else
  char ch = getch();

  if (ch == 27) {
    ch = getch();
    if (ch == 91) {
      ch = getch();
      switch (ch) {
      case 65:
	if (quanty >= YMIN) quanty--;
	break;
      case 66:
	if (quanty <= YMAX) quanty++;
	break;
      case 67:
	if (quantx <= XMAX) quantx++;
	break;
      case 68:
	if (quantx >= XMIN) quantx--;
	break;
      }
    }
  }
#endif
  *px = quantx;
  *py = quanty;

  delay(12);
}


void NewGame(void) {
  ClearScreen(VRAM);
  ClearScreen(map1);
  ClearScreen(map2);
  ClearScreen(map3);

  targetsize = 2;
  targethits = 0;
  totalhits = 0;
  maxhits = 0;

  playerhits = 0;
#ifdef __AVR__
  WaitCoin();
#endif
  Init();
}


void GetShipEndPoints(unsigned short ship, int * pxmin, int * pxmax, int * pymin, int * pymax)
{
  int xmin = ship & 0x000F;
  int ymin = (ship >> 4) & 0x000F;
  int xmax;
  int ymax;
  int size = (ship >> 14) + 2;

  if (ship & 0x2000) {
    xmax = xmin;
    ymax = ymin + size - 1;
  }
  else {
    ymax = ymin;
    xmax = xmin + size - 1;
  }

  *pxmin = xmin;
  *pxmax = xmax;
  *pymin = ymin;
  *pymax = ymax;
}


int collision(unsigned short ship1, unsigned short ship2)
{
  int dist = 0;

  int x1min, x1max, y1min, y1max;
  int x2min, x2max, y2min, y2max;


  GetShipEndPoints(ship1, &x1min, &x1max, &y1min, &y1max);
  GetShipEndPoints(ship2, &x2min, &x2max, &y2min, &y2max);


  if (x1max < x2min) {
    /* Schiff 1 links von Schiff 2 */
    dist = x2min - x1max;
  }
  else if (x1min > x2max) {
    /* Schiff 1 rechts von Schiff 2 */
    dist = x1min - x2max;
  }

  if (y1max < y2min) {
    /* Schiff 1 unterhalb von Schiff 2 */
    dist += y2min - y1max;
  }
  else if (y1min > y2max) {
    /* Schiff 1 oberhalb von Schiff 2 */
    dist += y1min - y2max;
  }

  return (dist < 2);
}


int CheckShipPos(unsigned short ship, int index)
{
  int first, i;

  int xmin, xmax, ymin, ymax;

  GetShipEndPoints(ship, &xmin, &xmax, &ymin, &ymax);

  if (xmin<XMIN || xmax>XMAX || ymin<YMIN || ymax>YMAX)
    return 1;

  if (index < SHIPS)
    first = 0;
  else
    first = SHIPS;

  for (i = first; i<index; i++) {
    if (collision(ships[i], ship))
      return 1;
  }

  return 0;
}


/*
  Einzelnes Schiff vom Spieler positionieren lassen
*/
void PutShip(int shipindex, int size)
{
  int x, y, xold, yold;
  int xmin, ymin, xmax, ymax;
  int vert = 0;
  int color;
  int timer = 250;
  unsigned short ship;

  GetXY(&xold, &yold);

  maxhits += size;

  /* Spielfeld ohne Schiff sichern */
  CopyScreen(VRAM, map1);

  while (timer > 0) {
    GetXY(&x, &y);

    /* Schiff an Spielfeldkanten drehen */
    if (x<XMIN || x>XMAX) {
      if (vert==0) {
	vert = 1;
	flipsound();
      }
    }
    else if (y<YMIN || y>YMAX) {
      if (vert) {
	vert = 0;
	flipsound();
      }
    }

    /* Extrema bestimmen */
    xmin = xmax = x;
    ymin = ymax = y;
    if (vert) {
      ymin -= size >> 1;
      ymax = ymin + size - 1;
    }
    else {
      xmin -= size >> 1;
      xmax = xmin + size - 1;
    }

    /* Spielfelduebertretung korrigieren */
    if (xmin < XMIN) {
      xmin = XMIN;
    }
    if (xmax > XMAX) {
      xmin = XMAX-(xmax-xmin);
    }
    if (ymin < YMIN) {
      ymin = YMIN;
    }
    if (ymax > YMAX) {
      ymin = YMAX-(ymax-ymin);
    }

    if (x==xold && y==yold) {
      timer--;
    }
    else {
      /* Positionsaenderung */
      xold = x;
      yold = y;
      timer = 200;
      click();
      /* Schiff an alter Position loeschen */
      CopyScreen(map1, VRAM);
    }

    TimeBar(timer);

    ship = ((size-2) << 14) | (vert << 13) | (ymin << 4) | xmin;
    ship |= ((1<<size) - 1) << 8; /* alle Segmente aktivieren */
 
    /* Schiff einzeichnen */
    if (CheckShipPos(ship, shipindex)) {
      /* Ungültige Position: blasse Darstellung */
      color = 1;
      if (timer<168)
	timer = 200;
      else {
	if (timer < 180) {
	  color = 0;
	}
      }
    }
    else {
      /* Gültige Position: helle Darstellung */
      color = 3;
      ships[shipindex] = ship;
    }

    DrawShip(ship, VRAM, color, 0);

  }

  /* Endgültige Position durch Blinken hervorheben' */
  for (timer=2; timer>0; timer--) {
    DrawShip(ship, VRAM, 0, 0);
    Beep(2000);
    DrawShip(ship, VRAM, 3, 0);
    Beep(1000);
    DrawShip(ship, VRAM, 1, 0);
    Beep(500);
  }
}


/*
  Fadenkreuz vom Spieler positionieren lassen
*/
void Aim(int *px, int *py)
{
  int x, y, xold, yold;
  int timer = 250;
  int radius = 1;

  GetXY(&xold, &yold);

  while (timer > 0) {

    GetXY(&x, &y);

    if (x<XMIN || x>XMAX) {
      /* Pause-Stellung */
      timer = 180;
    }
    /* Fadenkreuz auf Bildschirm begrenzen */
    if (x<0) x=0;
    if (x>11) x=11;
    if (y<YMIN) y=YMIN;
    if (y>YMAX) y=YMAX;

    if (x==xold && y==yold) {
      timer--;
      randomize();
    }
    else {
      /* Positionsaenderung */
      xold = x;
      yold = y;
      timer = 180;
      click();
      /* altes Fadenkreuz loeschen */
      CopyScreen(map2, VRAM);
    }

    TimeBar(timer);

    /* Fadenkreuz */
    {
      if (timer > 20) {
	if (radius < 10) {
	  radius += (timer & 1);
	}
      }
      else {
	/* altes Fadenkreuz loeschen */
	CopyScreen(map2, VRAM);
	radius -= (timer & 1);
      }

      int i;
      for (i=1; i<=radius; i++) {
	if (x-i >= XMIN)
	  MergePixel(x-i, y, map2);
	if (x+i <= XMAX)
	  MergePixel(x+i, y, map2);
	MergePixel(x, y-i, map2);
	MergePixel(x, y+i, map2);

      }
    }
  }

  /* Restore screen */
  {
    int i;
    for (i=SHIPS; i<2*SHIPS; i++)
      DrawShip(ships[i], map2, -1, 3);
  }
  CopyScreen(map2, VRAM);

  *px = x;
  *py = y;
}


void Explosion(int x, int y)
{
  int i, dx, dy, color;
#ifndef __AVR__
  dspfile = fopen("/dev/dsp", "wb");
#endif

  CopyScreen(VRAM, save);

  for (i=0; i<400; i++) {
    randomize();

    if ((randomstate & 0x8000) && (i < (randomstate & 511)))
      SpeakerOn(randomstate & 0x07FF);
    else
      SpeakerOff(randomstate & 0x07FF);

    /* Streuung zwischen -3 und +3 per Zufall bestimmen */
    dx =        (randomstate & 3) + ((randomstate >> 2) & 3) - 3;
    dy = ((randomstate >> 4) & 3) + ((randomstate >> 6) & 3) - 3;


    if (i > (randomstate >> 8)) {
      color = 0;
    }
    else 
      color = 3 - ((dx*dx + dy*dy) >> 3);

    if (i < 64 || (i > 192 && color>0)) {
      /* Kleinere Ausdehnung in der Anfangs- und Endphase */
      if (dx>2) dx=1;
      if (dx<-2) dx=-1;
      if (dy>2) dy=1;
      if (dy<-2) dy=-1;
    }

    if (color==0) {
      color = GetScreenPixel(x+dx, y+dy, save);
    }

    PutPixel(x+dx, y+dy, color);
  }

  CopyScreen(save, VRAM);

#ifndef __AVR__
  if (dspfile) {
    fclose(dspfile);
  }
#endif
}


int CheckHit(int x, int y, int startship, int stopship)
{
  int i, xmin, xmax, ymin, ymax, delta, size, color;

  FlashPixel(x, y);

  for (i=startship; i<stopship; i++) {
    unsigned short ship = ships[i];

    GetShipEndPoints(ship, &xmin, &xmax, &ymin, &ymax);
    size = (ship >> 14) + 2;

    if (ship & 0x2000) {
      delta = y - ymin;
      if (x != xmin)
	continue;
    }
    else {
      delta = x - xmin;
      if (y != ymin)
	continue;
    }

    if (delta >= 0 && delta < size) {
      /* Treffer => Schiffsegment löschen */
      int segment = 0x0100 << delta;
      if (ship & segment) {
	ship &= ~segment;
	ships[i] = ship;
      }
      else {
	/* Segment war schon zerstoert */
	break;
      }

      Explosion(x, y);

      if ((ship & 0x1F00) == 0) {
	int j;
	/* Schiff versenkt */
	for (j=0; j<7; j++) {
	  color = 1 + (j & 2);
	  if (ship & 0x2000)
	    VLine(xmin, ymin, size, color);
	  else
	    HLine(xmin, ymin, size, color);
	  Beep(5000);
	  delay(100);
	}
      }

      if (i<SHIPS) {
	/* Computer hat geschossen */
	targethits++;
	totalhits++;

	if (targethits==1) {
	  /* Unbekanntes Schiff getroffen => Neues Ziel setzen */
	  targetx = x;
	  targety = y;
	  distanz = 0;
	  flanke[0] = 0;
	  flanke[1] = 0;
	  flanke[2] = 0;
	  flanke[3] = 0;
	}
	if (targethits==2) {
	  /* Richtung liegt fest */
	  if (xmin==xmax) {
	    /* Schiff vertikal => Horizontale nicht weiter beachten */
	    flanke[0] = 10;
	    flanke[2] = 10;
	  }
	  else {
	    /* Schiff horizontal => Vertikale nicht weiter beachten */
	    flanke[1] = 10;
	    flanke[3] = 10;
	  }
	}

	if ((ship & 0x1F00) == 0) {
	  /* Computer hat Schiff versenkt */

	  /* Aktuelles Zielgebiet löschen */
	  targethits = 0;
	  targetx = -1;

	  /* Pufferzone kartieren */
	  if (xmin==xmax) {
	    /* vertikal */
	    VLineScreen(xmin-1, ymin, size, map3, 1);
	    VLineScreen(xmin+1, ymin, size, map3, 1);
	    PutScreenPixel(xmin, ymin-1, map3, 1);
	    PutScreenPixel(xmin, ymax+1, map3, 1);
	  }
	  else {
	    /* horizontal */
	    HLineScreen(xmin, ymin-1, size, map3, 1);
	    HLineScreen(xmin, ymin+1, size, map3, 1);
	    PutScreenPixel(xmin-1, ymin, map3, 1);
	    PutScreenPixel(xmin+1, ymin, map3, 1);
	  }
	}
      }
      else {
	/* Spieler hat geschossen */
	playerhits++;
      }
      return 1;
    }
  }

  /* Kein Treffer */
  for (i=0; i<3; i++) {
    FlashPixel(x, y);
  }
  PutPixel(x, y, 1);

  /* Kurze Pause nach Wassereinschlag */
  delay(500);

  return 0;
}


void Rakete(int x, int y, int dir)
{
  int newy, oldy, newcolor, oldcolor;
  unsigned char c;

  if (dir==1)
    newy = -1;
  else
    newy = 10;

  oldy = newy;

#ifndef __AVR__
  dspfile = fopen("/dev/dsp", "wb");
#endif

  for (c=17; c<200; c++) {
    int wait = (c*c) >> 3;
    randomize();

    if ((c & 15)==0) {
      oldy = newy;
      oldcolor = newcolor;
      newy = oldy + dir;
      newcolor = GetPixel(x, newy);
    }
    if (newy == y) {
      /* Ueber Ziel stehen bleiben */
      dir = 0;
    }

    PutPixel(x, oldy, 1);
    PutPixel(x, newy, 0);
    SpeakerOn(wait);

    PutPixel(x, oldy, 0);
    PutPixel(x, newy, 1);
    SpeakerOff(wait);

    PutPixel(x, oldy, oldcolor);
  }
#ifndef __AVR__
  if (dspfile) {
    fclose(dspfile);
  }
#endif
}


int Shoot(int x, int y)
{
  int i, hit;

  /* Kanonenschuss von unten nach oben */
  Rakete(x, y, -1);

  hit = CheckHit(x, y, SHIPS, 2*SHIPS);

  if (!hit) {
    PutScreenPixel(x, y, map2, 1); /* Wasser */
  }

  /* Restore screen */
  for (i=SHIPS; i<2*SHIPS; i++)
    DrawShip(ships[i], map2, -1, 3);

  CopyScreen(map2, VRAM);

  return hit;
}


void ComputerPutShip(int shipindex, int size)
{
  unsigned int vert;
  unsigned short ship;

  do {
    randomize();
      
    vert = randomstate & 0x2000;

    ship = ((size-2) << 14) | vert | (randomstate & 0x00FF);
    ship |= ((1<<size) - 1) << 8; /* alle Segmente aktivieren */
  }
  while (CheckShipPos(ship, shipindex));

  ships[shipindex] = ship;

  DrawShip(ship, map2, -1, 0); /* unsichtbar */
}


int CheckAim(int x, int y)
{
  int delta;
  int breite = 0;
  int hoehe = 0;

  if ( x < XMIN || x > XMAX || y < YMIN || y > YMAX ) {
    /* ausserhalb des Spielfelds */
    return 0;
  }

  /* Anzahl benachbarter ungetesteter Felder in jeder Richtung ermitteln */
  for (delta=0; delta<5; delta++) {
    if (x-delta < XMIN)
      break;
    if (GetScreenPixel(x-delta, y, map3) == 1)
      break;
    breite++;
  }
  if (breite) {
    for (delta=1; delta<5; delta++) {
      if (x+delta > XMAX)
	break;
      if (GetScreenPixel(x+delta, y, map3) == 1)
	break;
      breite++;
    }
  }
  for (delta=0; delta<5; delta++) {
    if (y-delta < YMIN)
      break;
    if (GetScreenPixel(x, y-delta, map3) == 1)
      break;
    hoehe++;
  }
  if (hoehe) {
    for (delta=1; delta<5; delta++) {
      if (y+delta > YMAX)
	break;
      if (GetScreenPixel(x, y+delta, map3) == 1)
	break;
      hoehe++;
    }
  }

  if (targethits) {
    /* Kein komplettem Schiff suchen, sondern nur 1 freies Feld */
    targetsize = 1;
  }

  return (breite >= targetsize || hoehe >= targetsize);
}


void ComputerAim(int *px, int *py)
{
  int x, y;
  int fire = 0;

  if (targethits) {
    /* Altes Ziel variieren */

    while (!fire) {
      randomize();
      richtung = randomstate & 3;
      distanz = flanke[richtung] + 1;
      x = targetx;
      y = targety;
      switch(richtung) {
      case 0:
	x += distanz;
	break;
      case 1:
	y += distanz;
	break;
      case 2:
	x -= distanz;
	break;
      case 3:
	y -= distanz;
	break;
      }
      fire = CheckAim(x, y);
    }
  }
  else {
    /* Neues Ziel suchen */
    while (!fire) {
      randomize();

      rasterx += 1 + (randomstate & 0x0007);
      rastery += 1 + (randomstate & 0x0007) + ((randomstate >> 4) & 2);
      if (rasterx > XMAX) rasterx -= (XMAX-XMIN+1);
      if (rastery > YMAX) rastery -= (YMAX-YMIN+1);
      fire = CheckAim(rasterx, rastery);
    }
    x = rasterx;
    y = rastery;
  }

  *px = x;
  *py = y;
}


int ComputerShot(int x, int y)
{
  int i, hit;

  /* Refresh screen */
  for (i=0; i<SHIPS; i++) {
    DrawShip(ships[i], map1, 3, 1);
  }
  CopyScreen(map1, VRAM);


  /* Kanonenschuss von oben nach unten */
  Rakete(x, y, 1);

  hit = CheckHit(x, y, 0, SHIPS);

  if (hit) {
    flanke[richtung] = distanz;
  }
  else {
    flanke[richtung] = 10; /* Richtung wird ungültig */
  }

  /* Computer-Karte */
  PutScreenPixel(x, y, map3, 1);

  /* Sichtbare Karte */
  PutScreenPixel(x, y, map1, 1);

  /* Schiffe neu zeichnen und kleinstes intaktes Schiff ermitteln */
  targetsize = 5;
  for (i=0; i<SHIPS; i++) {
    int uebrig = DrawShip(ships[i], map1, 3, 1);
    if (uebrig >= 2 && uebrig < targetsize) {
      targetsize = uebrig;
    }
  }

  CopyScreen(map1, VRAM);

  return hit;
}


void EndGame(int playerWins)
{
  int i,x;

  ClearScreen(save);
  if (playerWins) {
    ScrollUpScreen(save);
    VRAM[0][0] = 454;
    VRAM[0][1] = 536;
    VRAM[0][2] = 390;
    VRAM[0][3] = 512;
    VRAM[0][4] = 460;
    VRAM[0][5] = 18;
    VRAM[0][6] = 972;
    VRAM[0][7] = 0;
    VRAM[0][8] = 990;
    VRAM[0][9] = 144;
    VRAM[0][10] = 286;
    VRAM[0][11] = 960;
  }
  else {
    ScrollDownScreen(save);
    VRAM[0][0] = 448;
    VRAM[0][1] = 512;
    VRAM[0][2] = 384;
    VRAM[0][3] = 512;
    VRAM[0][4] = 448;
    VRAM[0][5] = 18;
    VRAM[0][6] = 990;
    VRAM[0][7] = 18;
    VRAM[0][8] = 960;
    VRAM[0][9] = 128;
    VRAM[0][10] = 256;
    VRAM[0][11] = 960;
  }

  for (i=0; i<32; i++) {
    int shift;
    if (playerWins)
      shift = ((i&15)>>2);
    else
      shift = ((i>>2) & 1);
    Beep(4000 >> shift);
  }
  for (i=0; i<4; i++) {
    for (x=0; x<12; x++) {
      VRAM[1][x] = VRAM[0][x];
    }
    delay(500);
    for (x=0; x<12; x++) {
      VRAM[1][x] = 0;
    }
    delay(500);
  }

  ClearScreen(VRAM);
  delay(2000);
}


int main (void) {
 
  int player, gameover;
  int map = 1;

  Init();

  while(1) {

    NewGame();

    /* Schiffe vom Spieler plazieren lassen */
    PutShip(0, 5);
    PutShip(1, 4);
    PutShip(2, 3);
    PutShip(3, 3);
    PutShip(4, 2);

    /* Schiffe des Computergegners aufstellen */
    ComputerPutShip(5, 5);
    ComputerPutShip(6, 4);
    ComputerPutShip(7, 3);
    ComputerPutShip(8, 3);
    ComputerPutShip(9, 2);

    player = randomstate & 1;
    gameover = 0;

    while (!gameover) {
      int x, y;

      if (player) {
	ScrollDownScreen(map2);
	map = 2;
      }
      while(player && !gameover) {
	Aim(&x, &y);
	player = Shoot(x, y);
	if (playerhits == maxhits) {
	  gameover = 1;
	}
      }

      if (map == 2 && !gameover) {
	ScrollUpScreen(map1);
	map = 1;
      }
      while (!player && !gameover) {
	ComputerAim(&x, &y);
	player = !ComputerShot(x, y);
	if (totalhits == maxhits)
	  gameover = 1;
      }
    }

    EndGame(playerhits == maxhits);

  }
  return 0;
}
